@@ -3,7 +3,8 @@ package org.ddahl.rscala
 import java.net._
 import java.io._
 import scala.language.dynamics
-import java.lang.ref.{Reference, PhantomReference, ReferenceQueue}
+import java.lang.ref.{Reference => JavaReference, PhantomReference, ReferenceQueue}
+import scala.sys.process.Process
 
 import server._
 import Protocol._
@@ -17,6 +18,8 @@ import Protocol._
 * In a Scala application, an instance of this class is created using its companion object.  See below.  The paths of the
 * rscala's JARs (for all supported versions of Scala) are available from [[http://www.r-project.org R]] using `rscala::.rscalaJar()`.
 * To get just the JAR for Scala 2.12, for example, use `rscala::.rscalaJar("2.12")`.
+*
+* This class is threadsafe.
 * 
 * {{{
 * val R = org.ddahl.rscala.RClient()
@@ -44,30 +47,53 @@ import Protocol._
 *
 * @author David B. Dahl
 */
-class RClient private (private val scalaServer: ScalaServer, private val in: DataInputStream, private val out: DataOutputStream, private val debugger: Debugger, val serializeOutput: Boolean, val rowMajor: Boolean) extends Dynamic {
+class RClient private (private val scalaServer: ScalaServer, private val rProcessInstance: Process, private val in: DataInputStream, private val out: DataOutputStream, private val debugger: Debugger, val serializeOutput: Boolean, val rowMajor: Boolean) extends Dynamic {
 
   import Helper.{readString, writeString, isMatrix, transposeIf}
 
-  private val referenceQueue = new ReferenceQueue[RPersistentReference]()
-  private val referenceMap = new scala.collection.mutable.HashMap[Reference[_ <: RPersistentReference],String]()
+  private val referenceQueue = new ReferenceQueue[PersistentReference]()
+  private val referenceMap = new scala.collection.mutable.HashMap[JavaReference[_ <: PersistentReference],String]()
 
   /** __For rscala developers only__: Returns `TRUE` if debugging output is enabled. */
   def debug = debugger.value
 
+  /** Pings the R interpreter.
+  *
+  * Calling this method periodically on an otherwise-idle client may prevent the operating system from closing the socket.
+  * Returns `true` if the ping is successful and `false` otherwise.
+  */
+  def ping(): Boolean = synchronized {
+    try {
+      out.writeInt(PING)
+      out.flush()
+      val status = in.readInt()
+      status == OK
+    } catch {
+      case _ : Throwable => false
+    }
+  }
+
   /** Closes the interface to the R interpreter.
   * 
   * Subsequent calls to the other methods will fail.
   */
-  def exit() = {
-    out.writeInt(SHUTDOWN)
-    out.flush()
+  def exit(): Unit = synchronized {
+    try {
+      check4GC()
+      out.writeInt(SHUTDOWN)
+      out.flush()
+    } catch {
+      case e : Throwable =>
+        if ( rProcessInstance != null ) rProcessInstance.destroy()
+        else throw e
+    }
   }
 
-  private def check4GC() {
+  private def check4GC() = {
     val first = referenceQueue.poll
     if ( first != null ) {
       out.writeInt(FREE)
-      var list = List[Reference[_ <: RPersistentReference]](first)
+      var list = List[JavaReference[_ <: PersistentReference]](first)
       while ( list.head != null ) {
         list = referenceQueue.poll :: list
       }
@@ -76,9 +102,7 @@ class RClient private (private val scalaServer: ScalaServer, private val in: Dat
       list.foreach( x => {
         val id = referenceMap(x)
         referenceMap.remove(x)
-        val index = id.indexOf("$")
-        val str = if ( index == -1 ) id else id.substring(index+1)
-        writeString(out,str)
+        writeString(out,id)
       })
       out.flush()
     }
@@ -86,13 +110,14 @@ class RClient private (private val scalaServer: ScalaServer, private val in: Dat
 
   /** Evaluates `snippet` in the R interpreter.
   *
-  * Returns `null` if `evalOnly`.  If `!evalOnly`, the last result of the R expression is converted if possible.
+  * Returns `null` if `evalOnly`.  If `!evalOnly`, the last result of the R expression is returned.  The result is
+  * converted if `asReference` is `false` and the conversion is supported, otherwise a reference is returned.
   * Conversion to integers, doubles, Booleans, and strings are supported, as are vectors (i.e. arrays) and matrices
   * (i.e. retangular arrays of arrays) of these types.  The static type of the result, however, is `Any` so using the
   * method `evalXY` (where `X` is `I`, `D`, `B`, `S`, or `R` and `Y` is `0`, `1`, or `2`) may be more convenient (e.g.
   * [[evalD0]]).
   */
-  def eval(snippet: String, evalOnly: Boolean): Any = {
+  def eval(snippet: String, evalOnly: Boolean, asReference: Boolean): (Any, String) = synchronized {
     check4GC()
     if ( debug ) debugger.msg("Sending EVAL request.")
     out.writeInt(if(serializeOutput) EVAL else EVALNAKED)
@@ -110,123 +135,196 @@ class RClient private (private val scalaServer: ScalaServer, private val in: Dat
       println(output)
     } else if ( debug ) debugger.msg("No output.")
     if ( status != OK ) throw new RuntimeException("Error in R evaluation.")
-    if ( evalOnly ) null else get(".rscala.last.value")._1
+    if ( evalOnly ) null else {
+      if ( debug ) debugger.msg("Getting EVAL result.")
+      if ( asReference ) out.writeInt(GET_REFERENCE) else out.writeInt(GET)
+      writeString(out,".rsX")
+      out.flush()
+      getInternal()
+    }
   }
 
   /** Calls '''`eval(snippet,true)`'''.  */
-  def eval(snippet: String): Unit = eval(snippet,true)
+  def eval(snippet: String): Unit = eval(snippet,true,false)
+
+  /** Calls '''`eval(snippet,true)`''' and returns the result using [[getReference]].  */
+  def evalReference(snippet: String) = toReference(eval(snippet,false,true))
 
   /** Calls '''`eval(snippet,true)`''' and returns the result using [[getI0]].  */
-  def evalI0(snippet: String) = { eval(snippet,true); getI0(".rscala.last.value") }
+  def evalI0(snippet: String) = toI0(eval(snippet,false,false))
 
   /** Calls '''`eval(snippet,true)`''' and returns the result using [[getD0]].  */
-  def evalD0(snippet: String) = { eval(snippet,true); getD0(".rscala.last.value") }
+  def evalD0(snippet: String) = toD0(eval(snippet,false,false))
 
   /** Calls '''`eval(snippet,true)`''' and returns the result using [[getL0]].  */
-  def evalL0(snippet: String) = { eval(snippet,true); getL0(".rscala.last.value") }
+  def evalL0(snippet: String) = toL0(eval(snippet,false,false))
 
   /** Calls '''`eval(snippet,true)`''' and returns the result using [[getS0]].  */
-  def evalS0(snippet: String) = { eval(snippet,true); getS0(".rscala.last.value") }
+  def evalS0(snippet: String) = toS0(eval(snippet,false,false))
 
   /** Calls '''`eval(snippet,true)`''' and returns the result using [[getR0]].  */
-  def evalR0(snippet: String) = { eval(snippet,true); getR0(".rscala.last.value") }
+  def evalR0(snippet: String) = toR0(eval(snippet,false,false))
 
   /** Calls '''`eval(snippet,true)`''' and returns the result using [[getI1]].  */
-  def evalI1(snippet: String) = { eval(snippet,true); getI1(".rscala.last.value") }
+  def evalI1(snippet: String) = toI1(eval(snippet,false,false))
 
   /** Calls '''`eval(snippet,true)`''' and returns the result using [[getD1]].  */
-  def evalD1(snippet: String) = { eval(snippet,true); getD1(".rscala.last.value") }
+  def evalD1(snippet: String) = toD1(eval(snippet,false,false))
 
   /** Calls '''`eval(snippet,true)`''' and returns the result using [[getL1]].  */
-  def evalL1(snippet: String) = { eval(snippet,true); getL1(".rscala.last.value") }
+  def evalL1(snippet: String) = toL1(eval(snippet,false,false))
 
   /** Calls '''`eval(snippet,true)`''' and returns the result using [[getS1]].  */
-  def evalS1(snippet: String) = { eval(snippet,true); getS1(".rscala.last.value") }
+  def evalS1(snippet: String) = toS1(eval(snippet,false,false))
 
   /** Calls '''`eval(snippet,true)`''' and returns the result using [[getR1]].  */
-  def evalR1(snippet: String) = { eval(snippet,true); getR1(".rscala.last.value") }
+  def evalR1(snippet: String) = toR1(eval(snippet,false,false))
 
   /** Calls '''`eval(snippet,true)`''' and returns the result using [[getI2]].  */
-  def evalI2(snippet: String) = { eval(snippet,true); getI2(".rscala.last.value") }
+  def evalI2(snippet: String) = toI2(eval(snippet,false,false))
 
   /** Calls '''`eval(snippet,true)`''' and returns the result using [[getD2]].  */
-  def evalD2(snippet: String) = { eval(snippet,true); getD2(".rscala.last.value") }
+  def evalD2(snippet: String) = toD2(eval(snippet,false,false))
 
   /** Calls '''`eval(snippet,true)`''' and returns the result using [[getL2]].  */
-  def evalL2(snippet: String) = { eval(snippet,true); getL2(".rscala.last.value") }
+  def evalL2(snippet: String) = toL2(eval(snippet,false,false))
 
   /** Calls '''`eval(snippet,true)`''' and returns the result using [[getS2]].  */
-  def evalS2(snippet: String) = { eval(snippet,true); getS2(".rscala.last.value") }
+  def evalS2(snippet: String) = toS2(eval(snippet,false,false))
 
   /** Calls '''`eval(snippet,true)`''' and returns the result using [[getR2]].  */
-  def evalR2(snippet: String) = { eval(snippet,true); getR2(".rscala.last.value") }
+  def evalR2(snippet: String) = toR2(eval(snippet,false,false))
 
   /** Invokes an R function with arguments.  */
-  def invoke(function: RReference, args: Any*) = eval(mkSnippet(function,args))
- 
+  def invoke(function: Reference, args: Any*): Unit = synchronized { eval(mkSnippet(function,args)) }
+
+  /** Invokes an R function with arguments.  */
+  def invoke(functionName: String, args: Any*): Unit = synchronized { eval(mkSnippet(functionName,args)) }
+
+  /** Invokes an R function with arguments and returns the result using [[getReference]].  */
+  def invokeReference(function: Reference, args: Any*) = synchronized { evalReference(mkSnippet(function,args)) }
+
+  /** Invokes an R function with arguments and returns the result using [[getReference]].  */
+  def invokeReference(functionName: String, args: Any*) = synchronized { evalReference(mkSnippet(functionName,args)) }
+
   /** Invokes an R function with arguments and returns the result using [[getI0]].  */
-  def invokeI0(function: RReference, args: Any*) = evalI0(mkSnippet(function,args))
+  def invokeI0(function: Reference, args: Any*) = synchronized { evalI0(mkSnippet(function,args)) }
+
+  /** Invokes an R function with arguments and returns the result using [[getI0]].  */
+  def invokeI0(functionName: String, args: Any*) = synchronized { evalI0(mkSnippet(functionName,args)) }
+
+  /** Invokes an R function with arguments and returns the result using [[getD0]].  */
+  def invokeD0(function: Reference, args: Any*) = synchronized { evalD0(mkSnippet(function,args)) }
   
   /** Invokes an R function with arguments and returns the result using [[getD0]].  */
-  def invokeD0(function: RReference, args: Any*) = evalD0(mkSnippet(function,args))
+  def invokeD0(functionName: String, args: Any*) = synchronized { evalD0(mkSnippet(functionName,args)) }
   
   /** Invokes an R function with arguments and returns the result using [[getL0]].  */
-  def invokeL0(function: RReference, args: Any*) = evalL0(mkSnippet(function,args))
+  def invokeL0(function: Reference, args: Any*) = synchronized { evalL0(mkSnippet(function,args)) }
+  
+  /** Invokes an R function with arguments and returns the result using [[getL0]].  */
+  def invokeL0(functionName: String, args: Any*) = synchronized { evalL0(mkSnippet(functionName,args)) }
   
   /** Invokes an R function with arguments and returns the result using [[getS0]].  */
-  def invokeS0(function: RReference, args: Any*) = evalS0(mkSnippet(function,args))
+  def invokeS0(function: Reference, args: Any*) = synchronized { evalS0(mkSnippet(function,args)) }
+  
+  /** Invokes an R function with arguments and returns the result using [[getS0]].  */
+  def invokeS0(functionName: String, args: Any*) = synchronized { evalS0(mkSnippet(functionName,args)) }
   
   /** Invokes an R function with arguments and returns the result using [[getR0]].  */
-  def invokeR0(function: RReference, args: Any*) = evalR0(mkSnippet(function,args))
+  def invokeR0(function: Reference, args: Any*) = synchronized { evalR0(mkSnippet(function,args)) }
+  
+  /** Invokes an R function with arguments and returns the result using [[getR0]].  */
+  def invokeR0(functionName: String, args: Any*) = synchronized { evalR0(mkSnippet(functionName,args)) }
   
   /** Invokes an R function with arguments and returns the result using [[getI1]].  */
-  def invokeI1(function: RReference, args: Any*) = evalI1(mkSnippet(function,args))
+  def invokeI1(function: Reference, args: Any*) = synchronized { evalI1(mkSnippet(function,args)) }
+  
+  /** Invokes an R function with arguments and returns the result using [[getI1]].  */
+  def invokeI1(functionName: String, args: Any*) = synchronized { evalI1(mkSnippet(functionName,args)) }
   
   /** Invokes an R function with arguments and returns the result using [[getD1]].  */
-  def invokeD1(function: RReference, args: Any*) = evalD1(mkSnippet(function,args))
+  def invokeD1(function: Reference, args: Any*) = synchronized { evalD1(mkSnippet(function,args)) }
+  
+  /** Invokes an R function with arguments and returns the result using [[getD1]].  */
+  def invokeD1(functionName: String, args: Any*) = synchronized { evalD1(mkSnippet(functionName,args)) }
   
   /** Invokes an R function with arguments and returns the result using [[getL1]].  */
-  def invokeL1(function: RReference, args: Any*) = evalL1(mkSnippet(function,args))
+  def invokeL1(function: Reference, args: Any*) = synchronized { evalL1(mkSnippet(function,args)) }
+  
+  /** Invokes an R function with arguments and returns the result using [[getL1]].  */
+  def invokeL1(functionName: String, args: Any*) = synchronized { evalL1(mkSnippet(functionName,args)) }
   
   /** Invokes an R function with arguments and returns the result using [[getS1]].  */
-  def invokeS1(function: RReference, args: Any*) = evalS1(mkSnippet(function,args))
+  def invokeS1(function: Reference, args: Any*) = synchronized { evalS1(mkSnippet(function,args)) }
+  
+  /** Invokes an R function with arguments and returns the result using [[getS1]].  */
+  def invokeS1(functionName: String, args: Any*) = synchronized { evalS1(mkSnippet(functionName,args)) }
   
   /** Invokes an R function with arguments and returns the result using [[getR1]].  */
-  def invokeR1(function: RReference, args: Any*) = evalR1(mkSnippet(function,args))
+  def invokeR1(function: Reference, args: Any*) = synchronized { evalR1(mkSnippet(function,args)) }
+  
+  /** Invokes an R function with arguments and returns the result using [[getR1]].  */
+  def invokeR1(functionName: String, args: Any*) = synchronized { evalR1(mkSnippet(functionName,args)) }
   
   /** Invokes an R function with arguments and returns the result using [[getI2]].  */
-  def invokeI2(function: RReference, args: Any*) = evalI2(mkSnippet(function,args))
+  def invokeI2(function: Reference, args: Any*) = synchronized { evalI2(mkSnippet(function,args)) }
+  
+  /** Invokes an R function with arguments and returns the result using [[getI2]].  */
+  def invokeI2(functionName: String, args: Any*) = synchronized { evalI2(mkSnippet(functionName,args)) }
   
   /** Invokes an R function with arguments and returns the result using [[getD2]].  */
-  def invokeD2(function: RReference, args: Any*) = evalD2(mkSnippet(function,args))
+  def invokeD2(function: Reference, args: Any*) = synchronized { evalD2(mkSnippet(function,args)) }
+  
+  /** Invokes an R function with arguments and returns the result using [[getD2]].  */
+  def invokeD2(functionName: String, args: Any*) = synchronized { evalD2(mkSnippet(functionName,args)) }
   
   /** Invokes an R function with arguments and returns the result using [[getL2]].  */
-  def invokeL2(function: RReference, args: Any*) = evalL2(mkSnippet(function,args))
+  def invokeL2(function: Reference, args: Any*) = synchronized { evalL2(mkSnippet(function,args)) }
+  
+  /** Invokes an R function with arguments and returns the result using [[getL2]].  */
+  def invokeL2(functionName: String, args: Any*) = synchronized { evalL2(mkSnippet(functionName,args)) }
   
   /** Invokes an R function with arguments and returns the result using [[getS2]].  */
-  def invokeS2(function: RReference, args: Any*) = evalS2(mkSnippet(function,args))
+  def invokeS2(function: Reference, args: Any*) = synchronized { evalS2(mkSnippet(function,args)) }
+  
+  /** Invokes an R function with arguments and returns the result using [[getS2]].  */
+  def invokeS2(functionName: String, args: Any*) = synchronized { evalS2(mkSnippet(functionName,args)) }
   
   /** Invokes an R function with arguments and returns the result using [[getR2]].  */
-  def invokeR2(function: RReference, args: Any*) = evalR2(mkSnippet(function,args))
+  def invokeR2(function: Reference, args: Any*) = synchronized { evalR2(mkSnippet(function,args)) }
   
-  private def mkSnippet(function: RReference, args: Seq[Any]) = {
+  /** Invokes an R function with arguments and returns the result using [[getR2]].  */
+  def invokeR2(functionName: String, args: Any*) = synchronized { evalR2(mkSnippet(functionName,args)) }
+  
+  private def mkSnippet(functionName: String, args: Seq[Any]): String = mkSnippet(EphemeralReference(functionName), args)
+
+  private def mkSnippet(function: Reference, args: Seq[Any]): String = {
     var counter = 0
     val argsStrings = args.map {
       case null => "NULL"
-      case (name: String, r: RReference) => name + "=" + r.toString
+      case (name: String, r: PersistentReference) =>  s"$name = get('$r',envir=.rsI[['r']])"
+      case (name: String, r: EphemeralReference) =>  s"$name = $r"
       case (name: String, o) =>
         val id = ".rsX" + counter
         counter += 1
         set(id,o)
         name + "=" + id
-      case r: RReference => r.toString
+      case r: PersistentReference => s"get('$r',envir=.rsI[['r']])"
+      case r: EphemeralReference => r.toString
       case o =>
         val id = ".rsX" + counter
         counter += 1
         set(id,o)
         id
     }
-    function + "(" + argsStrings.mkString(",") + ")"
+    val functionString = function match {
+      case r: PersistentReference => s"get('$r',envir=.rsI[['r']])"
+      case _ => function.toString
+    }
+    val snippet = functionString + "(" + argsStrings.mkString(",") +")"
+    if ( debug ) debugger.msg("Constructed R snippet: "+snippet)
+    snippet
   }
 
   /** A short-hand way to call [[get]].
@@ -277,7 +375,7 @@ class RClient private (private val scalaServer: ScalaServer, private val in: Dat
   * R.eval("print(myList)")
   * }}}
   */
-  def set(identifier: String, value: Any, index: String = "", singleBrackets: Boolean = true): Unit = {
+  def set(identifier: String, value: Any, index: String = "", singleBrackets: Boolean = true): Unit = synchronized {
     check4GC()
     if ( debug ) debugger.msg("Setting: "+identifier)
     val v = value
@@ -472,11 +570,11 @@ class RClient private (private val scalaServer: ScalaServer, private val in: Dat
   }
 
   /** __For rscala developers only__:  Returns a value previously cached for the R interpreter. */
-  def cached(identifier: String): Any = {
+  def cached(identifier: String): Any = synchronized {
     if ( identifier.startsWith(".") ) {
       scalaServer.cacheMap(identifier)._1
     } else {
-      scalaServer.repl.valueOfTerm(identifier).get
+      scalaServer.repl.valueOfTerm(identifier).getOrElse(null)
     }
   }
 
@@ -487,12 +585,17 @@ class RClient private (private val scalaServer: ScalaServer, private val in: Dat
   * arrays) and matrices (i.e. retangular arrays of arrays) of these types.    Using the method `getXY` (where `X` is
   * `I`, `D`, `B`, or `S` and `Y` is `0`, `1`, or `2`) may be more convenient (e.g.  [[getD0]]).
   */
-  def get(identifier: String, asReference: Boolean = false): (Any,String) = {
+  def get(identifier: String, asReference: Boolean = false): (Any, String) = synchronized {
     check4GC()
     if ( debug ) debugger.msg("Getting: "+identifier)
     if ( asReference ) out.writeInt(GET_REFERENCE) else out.writeInt(GET)
     writeString(out,identifier)
     out.flush()
+    getInternal()
+  }
+
+  private def getInternal(): (Any, String) = {
+    if ( debug ) debugger.msg("Getting internal.")
     in.readInt match {
       case NULLTYPE =>
         if ( debug ) debugger.msg("Getting null.")
@@ -534,25 +637,32 @@ class RClient private (private val scalaServer: ScalaServer, private val in: Dat
         }
       case REFERENCE =>
         if ( debug ) debugger.msg("Getting reference.")
-        val reference = RPersistentReference(readString(in))
+        val reference = PersistentReference(readString(in))
         val phantomReference = new PhantomReference(reference, referenceQueue)
         referenceMap(phantomReference) = reference.name
-        (reference, "org.ddahl.rscala.RPersistentReference")
-      case UNDEFINED_IDENTIFIER => throw new RuntimeException("Undefined identifier: "+identifier)
-      case UNSUPPORTED_STRUCTURE => throw new RuntimeException("Unsupported data type: "+identifier)
+        (reference, "org.ddahl.rscala.PersistentReference")
+      case UNDEFINED_IDENTIFIER => throw new RuntimeException("Undefined identifier.")
+      case UNSUPPORTED_STRUCTURE => throw new RuntimeException("Unsupported data type.")
       case e => throw new RuntimeException("Protocol error: Unknown type: "+e)
     }
   }
 
+  /** Obtains a persistent R reference to the named object so that it can be accessed outside of the current environment. */
+  def getReference(identifier: String): PersistentReference = toReference(get(identifier,true))
+
   /** Converts an ephemeral R reference to a persistent R reference so that it can be accessed outside of the current environment. */
-  def makePersistent(reference: REphemeralReference): RPersistentReference = get(reference.name,true)._1.asInstanceOf[RPersistentReference]
+  def getReference(reference: EphemeralReference): PersistentReference = getReference(reference.name)
+
+  private def toReference(x: (Any, String)): PersistentReference = x._1.asInstanceOf[PersistentReference]
 
   /** Calls '''`get(identifier,false)`''' and converts the result to an `Int`.
   *
   * Integers, doubles, Booleans, and strings are supported.  Vectors (i.e. arrays) of these types are also supported by
   * converting the first element.  Matrices (i.e. rectangular arrays of arrays) are not supported.
   */
-  def getI0(identifier: String): Int = get(identifier) match {
+  def getI0(identifier: String): Int = toI0(get(identifier))
+
+  private def toI0(x: (Any, String)): Int = x match {
     case (a,"Int") => a.asInstanceOf[Int]
     case (a,"Double") => a.asInstanceOf[Double].toInt
     case (a,"Boolean") => if (a.asInstanceOf[Boolean]) 1 else 0
@@ -571,7 +681,9 @@ class RClient private (private val scalaServer: ScalaServer, private val in: Dat
   * Integers, doubles, Booleans, and strings are supported.  Vectors (i.e. arrays) of these types are also supported by
   * converting the first element.  Matrices (i.e. rectangular arrays of arrays) are not supported.
   */
-  def getD0(identifier: String): Double = get(identifier) match {
+  def getD0(identifier: String): Double = toD0(get(identifier))
+
+  private def toD0(x: (Any, String)): Double = x match {
     case (a,"Int") => a.asInstanceOf[Int].toDouble
     case (a,"Double") => a.asInstanceOf[Double]
     case (a,"Boolean") => if (a.asInstanceOf[Boolean]) 1.0 else 0.0
@@ -590,7 +702,9 @@ class RClient private (private val scalaServer: ScalaServer, private val in: Dat
   * Integers, doubles, Booleans, and strings are supported.  Vectors (i.e. arrays) of these types are also supported by
   * converting the first element.  Matrices (i.e. rectangular arrays of arrays) are not supported.
   */
-  def getL0(identifier: String): Boolean = get(identifier) match {
+  def getL0(identifier: String): Boolean = toL0(get(identifier))
+
+  private def toL0(x: (Any, String)): Boolean = x match {
     case (a,"Int") => a.asInstanceOf[Int] != 0
     case (a,"Double") => a.asInstanceOf[Double] != 0.0
     case (a,"Boolean") => a.asInstanceOf[Boolean]
@@ -609,7 +723,9 @@ class RClient private (private val scalaServer: ScalaServer, private val in: Dat
   * Integers, doubles, Booleans, and strings are supported.  Vectors (i.e. arrays) of these types are also supported by
   * converting the first element.  Matrices (i.e. rectangular arrays of arrays) are not supported.
   */
-  def getS0(identifier: String): String = get(identifier) match {
+  def getS0(identifier: String): String = toS0(get(identifier))
+
+  private def toS0(x: (Any, String)): String = x match {
     case (a,"Int") => a.asInstanceOf[Int].toString
     case (a,"Double") => a.asInstanceOf[Double].toString
     case (a,"Boolean") => a.asInstanceOf[Boolean].toString
@@ -628,7 +744,9 @@ class RClient private (private val scalaServer: ScalaServer, private val in: Dat
   * Integers, doubles, Booleans, and strings are supported.  Vectors (i.e. arrays) of these types are also supported by
   * converting the first element.  Matrices (i.e. rectangular arrays of arrays) are not supported.
   */
-  def getR0(identifier: String): Byte = get(identifier) match {
+  def getR0(identifier: String): Byte = toR0(get(identifier))
+
+  private def toR0(x: (Any, String)): Byte = x match {
     case (a,"Int") => a.asInstanceOf[Int].toByte
     case (a,"Double") => a.asInstanceOf[Double].toByte
     case (a,"Boolean") => if (a.asInstanceOf[Boolean]) 1.toByte else 0.toByte
@@ -647,7 +765,9 @@ class RClient private (private val scalaServer: ScalaServer, private val in: Dat
   * Integers, doubles, Booleans, and strings are supported.  Vectors (i.e. arrays) of these types are also supported by
   * converting the first element.  Matrices (i.e. rectangular arrays of arrays) are not supported.
   */
-  def getI1(identifier: String): Array[Int] = get(identifier) match {
+  def getI1(identifier: String): Array[Int] = toI1(get(identifier))
+
+  private def toI1(x: (Any, String)): Array[Int] = x match {
     case (a,"Int") => Array(a.asInstanceOf[Int])
     case (a,"Double") => Array(a.asInstanceOf[Double].toInt)
     case (a,"Boolean") => Array(if (a.asInstanceOf[Boolean]) 1 else 0)
@@ -666,7 +786,9 @@ class RClient private (private val scalaServer: ScalaServer, private val in: Dat
   * Integers, doubles, Booleans, and strings are supported.  Vectors (i.e. arrays) of these types are also supported by
   * converting the first element.  Matrices (i.e. rectangular arrays of arrays) are not supported.
   */
-  def getD1(identifier: String): Array[Double] = get(identifier) match {
+  def getD1(identifier: String): Array[Double] = toD1(get(identifier))
+
+  private def toD1(x: (Any, String)): Array[Double] = x match {
     case (a,"Int") => Array(a.asInstanceOf[Int].toDouble)
     case (a,"Double") => Array(a.asInstanceOf[Double])
     case (a,"Boolean") => Array(if (a.asInstanceOf[Boolean]) 1.0 else 0.0)
@@ -685,7 +807,9 @@ class RClient private (private val scalaServer: ScalaServer, private val in: Dat
   * Integers, doubles, Booleans, and strings are supported.  Vectors (i.e. arrays) of these types are also supported by
   * converting the first element.  Matrices (i.e. rectangular arrays of arrays) are not supported.
   */
-  def getL1(identifier: String): Array[Boolean] = get(identifier) match {
+  def getL1(identifier: String): Array[Boolean] = toL1(get(identifier))
+
+  private def toL1(x: (Any, String)): Array[Boolean] = x match {
     case (a,"Int") => Array(a.asInstanceOf[Int] != 0)
     case (a,"Double") => Array(a.asInstanceOf[Double] != 0.0)
     case (a,"Boolean") => Array(a.asInstanceOf[Boolean])
@@ -704,7 +828,9 @@ class RClient private (private val scalaServer: ScalaServer, private val in: Dat
   * Integers, doubles, Booleans, and strings are supported.  Vectors (i.e. arrays) of these types are also supported by
   * converting the first element.  Matrices (i.e. rectangular arrays of arrays) are not supported.
   */
-  def getS1(identifier: String): Array[String] = get(identifier) match {
+  def getS1(identifier: String): Array[String] = toS1(get(identifier))
+
+  private def toS1(x: (Any, String)): Array[String] = x match {
     case (a,"Int") => Array(a.asInstanceOf[Int].toString)
     case (a,"Double") => Array(a.asInstanceOf[Double].toString)
     case (a,"Boolean") => Array(a.asInstanceOf[Boolean].toString)
@@ -723,7 +849,9 @@ class RClient private (private val scalaServer: ScalaServer, private val in: Dat
   * Integers, doubles, Booleans, and strings are supported.  Vectors (i.e. arrays) of these types are also supported by
   * converting the first element.  Matrices (i.e. rectangular arrays of arrays) are not supported.
   */
-  def getR1(identifier: String): Array[Byte] = get(identifier) match {
+  def getR1(identifier: String): Array[Byte] = toR1(get(identifier))
+
+  private def toR1(x: (Any, String)): Array[Byte] = x match {
     case (a,"Int") => Array(a.asInstanceOf[Int].toByte)
     case (a,"Double") => Array(a.asInstanceOf[Double].toByte)
     case (a,"Boolean") => Array(if (a.asInstanceOf[Boolean]) 1.toByte else 0.toByte)
@@ -743,7 +871,9 @@ class RClient private (private val scalaServer: ScalaServer, private val in: Dat
   * Booleans, and strings themselves are not supported.  Vectors (i.e. arrays) of these
   * types are also not supported.
   */
-  def getI2(identifier: String): Array[Array[Int]] = get(identifier) match {
+  def getI2(identifier: String): Array[Array[Int]] = toI2(get(identifier))
+
+  private def toI2(x: (Any, String)): Array[Array[Int]] = x match {
     case (a,"Array[Array[Int]]") => a.asInstanceOf[Array[Array[Int]]]
     case (a,"Array[Array[Double]]") => a.asInstanceOf[Array[Array[Double]]].map(_.map(_.toInt))
     case (a,"Array[Array[Boolean]]") => a.asInstanceOf[Array[Array[Boolean]]].map(_.map(x => if (x) 1 else 0))
@@ -758,7 +888,9 @@ class RClient private (private val scalaServer: ScalaServer, private val in: Dat
   * Booleans, and strings themselves are not supported.  Vectors (i.e. arrays) of these
   * types are also not supported.
   */
-  def getD2(identifier: String): Array[Array[Double]] = get(identifier) match {
+  def getD2(identifier: String): Array[Array[Double]] = toD2(get(identifier))
+
+  private def toD2(x: (Any, String)): Array[Array[Double]] = x match {
     case (a,"Array[Array[Int]]") => a.asInstanceOf[Array[Array[Int]]].map(_.map(_.toDouble))
     case (a,"Array[Array[Double]]") => a.asInstanceOf[Array[Array[Double]]]
     case (a,"Array[Array[Boolean]]") => a.asInstanceOf[Array[Array[Boolean]]].map(_.map(x => if (x) 1.0 else 0.0))
@@ -773,7 +905,9 @@ class RClient private (private val scalaServer: ScalaServer, private val in: Dat
   * Booleans, and strings themselves are not supported.  Vectors (i.e. arrays) of these
   * types are also not supported.
   */
-  def getL2(identifier: String): Array[Array[Boolean]] = get(identifier) match {
+  def getL2(identifier: String): Array[Array[Boolean]] = toL2(get(identifier))
+
+  private def toL2(x: (Any, String)): Array[Array[Boolean]] = x match {
     case (a,"Array[Array[Int]]") => a.asInstanceOf[Array[Array[Int]]].map(_.map(_ != 0))
     case (a,"Array[Array[Double]]") => a.asInstanceOf[Array[Array[Double]]].map(_.map(_ != 0.0))
     case (a,"Array[Array[Boolean]]") => a.asInstanceOf[Array[Array[Boolean]]]
@@ -788,7 +922,9 @@ class RClient private (private val scalaServer: ScalaServer, private val in: Dat
   * Booleans, and strings themselves are not supported.  Vectors (i.e. arrays) of these
   * types are also not supported.
   */
-  def getS2(identifier: String): Array[Array[String]] = get(identifier) match {
+  def getS2(identifier: String): Array[Array[String]] = toS2(get(identifier))
+
+  private def toS2(x: (Any, String)): Array[Array[String]] = x match {
     case (a,"Array[Array[Int]]") => a.asInstanceOf[Array[Array[Int]]].map(_.map(_.toString))
     case (a,"Array[Array[Double]]") => a.asInstanceOf[Array[Array[Double]]].map(_.map(_.toString))
     case (a,"Array[Array[Boolean]]") => a.asInstanceOf[Array[Array[Boolean]]].map(_.map(_.toString))
@@ -803,7 +939,9 @@ class RClient private (private val scalaServer: ScalaServer, private val in: Dat
   * Booleans, and strings themselves are not supported.  Vectors (i.e. arrays) of these
   * types are also not supported.
   */
-  def getR2(identifier: String): Array[Array[Byte]] = get(identifier) match {
+  def getR2(identifier: String): Array[Array[Byte]] = toR2(get(identifier))
+
+  private def toR2(x: (Any, String)): Array[Array[Byte]] = x match {
     case (a,"Array[Array[Int]]") => a.asInstanceOf[Array[Array[Int]]].map(_.map(_.toByte))
     case (a,"Array[Array[Double]]") => a.asInstanceOf[Array[Array[Double]]].map(_.map(_.toByte))
     case (a,"Array[Array[Boolean]]") => a.asInstanceOf[Array[Array[Boolean]]].map(_.map(x => if (x) 1.toByte else 0.toByte))
@@ -830,13 +968,7 @@ object RClient {
 
   import scala.sys.process._
 
-  private val isWindows: Boolean = sys.props("os.name").toLowerCase match {
-    case s if s.startsWith("""windows""") => true
-    case s if s.startsWith("""linux""") => false
-    case s if s.startsWith("""unix""") => false
-    case s if s.startsWith("""mac""") => false
-    case _ => throw new RuntimeException("Unrecognized OS")
-  }
+  private val isWindows = scala.util.Properties.isWin
 
   private val defaultArguments = isWindows match {
     case true  => Array[String]("--no-save","--no-restore","--silent","--slave") 
@@ -916,7 +1048,7 @@ object RClient {
   * should be serialized, whether matrices are row major, whether debugging output should be displayed, the `timeout` to establish a connection
   * with the R interpreter, and the port number.  Two sockets are establised using: 1. the specified port and 2. the specified port plus one.
   */
-  def apply(rCmd: String = defaultRCmd, serializeOutput: Boolean = false, rowMajor: Boolean = true, port: Int = 0, debug: Boolean = false, timeout: Int = 60): RClient = {
+  def apply(rCmd: String = defaultRCmd, serializeOutput: Boolean = true, rowMajor: Boolean = true, port: Int = 0, debug: Boolean = false, timeout: Int = 60): RClient = {
     var cmd: PrintWriter = null
     val command = rCmd +: ( defaultArguments ++ interactiveArguments )
     val processCmd = Process(command)
@@ -927,7 +1059,7 @@ object RClient {
       reader(debugger,""),
       true
     )
-    val processInstance = processCmd.run(processIO)
+    val rProcessInstance = processCmd.run(processIO)
     val codeInR = Seq("common.R","globals.R","protocol.R","rServer.R","scala.R","zzz.R").map(resource => {
       scala.io.Source.fromInputStream(getClass.getResourceAsStream("/R/"+resource)).getLines.mkString("\n")
     }).mkString("\n\n")
@@ -956,11 +1088,11 @@ object RClient {
     val sockets = new ScalaSockets(portsFile.getAbsolutePath,port,debugger)
     sockets.out.writeInt(OK)
     sockets.out.flush()
-    apply(null,sockets.in,sockets.out,debugger,serializeOutput,rowMajor)
+    apply(null,rProcessInstance,sockets.in,sockets.out,debugger,serializeOutput,rowMajor)
   }
 
   /** __For rscala developers only__: Returns an instance of the [[RClient]] class.  */
-  def apply(scalaServer: ScalaServer, in: DataInputStream, out: DataOutputStream, debugger: Debugger, serializeOutput: Boolean, rowMajor: Boolean): RClient = new RClient(scalaServer,in,out,debugger,serializeOutput,rowMajor)
+  def apply(scalaServer: ScalaServer, rProcessInstance: Process, in: DataInputStream, out: DataOutputStream, debugger: Debugger, serializeOutput: Boolean, rowMajor: Boolean): RClient = new RClient(scalaServer,rProcessInstance,in,out,debugger,serializeOutput,rowMajor)
 
 }
 
